
from barviz.Simplex        import Simplex
from barviz.SimplexBuilder import SimplexBuilder
from barviz.Scrawler       import Scrawler
from barviz.Attributes     import Attributes
from barviz.Collection     import Collection
from barviz.version        import __version__

version = __version__





# ---------------------------------------------------------
#    ____          __      ___     
#   |  _ \         \ \    / (_)    
#   | |_) | __ _ _ _\ \  / / _ ____
#   |  _ < / _` | '__\ \/ / | |_  /
#   | |_) | (_| | |   \  /  | |/ / 
#   |____/ \__,_|_|    \/   |_/___|  SimplexBuilder class
# ---------------------------------------------------------                            
# SIMaP / Jean-Luc Parouty 2022                  

import numpy as np
import math, random
from math import sqrt, cos, sin, acos

from barviz.version        import __version__


class SimplexBuilder:

    version = __version__

    def __init__(self, m=10, q=10, ks=10, s0=2, kd=4):
        self.m    = m        # mass
        self.q    = q        # charge        
        self.ks   = ks       # spring coefficient
        self.s0   = s0       # spring elongation at rest
        self.kd   = kd       # damping factor
        self.ee   = 0.001    # epsilon to avoid /0


    def add_points(self, nbp=10, mode='fibonacci'):
        self.__set_random_uniform_points(nbp)
        if mode=='fibonacci': self.__set_fibonacci_points(nbp)
        if mode=='spherical': self.__set_random_spherical_points(nbp)


    @property
    def nbp(self):
        return len(self.points)


    def get_simplex(self):
        # Import here, to avoid circular reference..
        from barviz.Simplex import Simplex
        nbp=self.nbp
        return Simplex( name     = f'{nbp-1}-simplex', 
                        points   = self.points,
                        labels   = None,
                        colors   = None ) 


    def iterate(self, max_dp=0.01, max_i=5000, dt=0.01, epsilon=0.001):
        done      = False
        iteration = 0
        p         = self.points           # points
        nbp       = self.nbp              # nb points
        dp        = np.zeros( (nbp,3) )   # speeds

        trace_dp  = []

        while done is False:

            # ---- Force : Spring to origin / F = -ks * ( d - s0 )

            fs = np.zeros((nbp,3))
            for i in range(nbp):
                pi=p[i]
                di = np.sqrt(pi.dot(pi))
                fs[i] = -self.ks * ( di - self.s0 ) * pi

            # ---- Force : Electric repulsion / Fij = q * ij / d*d*d
            #
            fq = np.zeros((nbp,3))
            for i in range(nbp):
                for j in range(nbp):
                    if i==j : next
                    ij     = p[j] - p[i]
                    d      = np.sqrt(ij.dot(ij))            
                    fq[i] += -self.q * ij / ( d*d*d + epsilon )
                    
            # ---- Force : Damping / F = -kd * dp
            #
            fd = -self.kd * dp

            # ---- Fine, now, we can update our points / F=m.a => a=F/m => dp += (F/m).dt
            #
            dp += ( fs + fq + fd) * dt / self.m

            p += dp * dt

            # ---- Sum of the speeds 
            sum_dp=0
            for dpi in dp:
                sum_dp+= np.sqrt(dpi.dot(dpi) )
            trace_dp.append(sum_dp)

            # ---- Stop condition
            #
            iteration += 1
            if iteration>max_i or sum_dp<max_dp: done=True

        self.trace_dp = trace_dp
        return sum_dp, iteration


    def __set_fibonacci_points(self, nbp=10):
        points = []
        phi = math.pi * (3. - sqrt(5.))
        for i in range(nbp):
            y = 1 - (i / float(nbp - 1)) * 2
            radius = sqrt(1 - y * y)
            theta = phi * i
            x = cos(theta) * radius
            z = sin(theta) * radius
            points.append((x, y, z))
        self.points = np.array(points)

    def __set_random_uniform_points(self, nbp=10):
        self.points = np.random.uniform(low=-1, high=1, size=(nbp,3))


    def __set_random_spherical_points(self, nbp=10):
        points=[]
        for i in range(nbp):
            theta = 2 * math.pi * random.random()
            phi = acos(random.uniform(-1, 1))
            x = cos(theta) * sin(phi)
            y = sin(theta) * sin(phi)
            z = cos(phi)
            points.append((x, y, z))
        self.points = np.array(points)




# ---------------------------------------------------------
#    ____          __      ___     
#   |  _ \         \ \    / (_)    
#   | |_) | __ _ _ _\ \  / / _ ____
#   |  _ < / _` | '__\ \/ / | |_  /
#   | |_) | (_| | |   \  /  | |/ / 
#   |____/ \__,_|_|    \/   |_/___|  Scrawler class
# ---------------------------------------------------------                            
# SIMaP / Jean-Luc Parouty 2022    

from barviz.Collection import Collection
import numpy as np

import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio

from barviz.version        import __version__

import random
from collections.abc import Iterable


class Scrawler:

    version = __version__

    def __init__(self, simplex=None):
        self.simplex    = simplex
        self.fig        = None

    def barycentric_to_3d(self,collection):
        vertices = self.simplex.points
        points   = collection.points
        nbp      = collection.nbp

        p3d = np.zeros( (nbp,3) )

        for i,p in enumerate(points):
            if None in p:
                p3d[i] = np.array([None,None,None])
            else:
                p3d[i] = p.dot(vertices)

        return p3d


    def update_center(self, observed_point):
        '''
        Get an observable point and update the plotly camera center position
        args :
            observable_point : point or collection
        return:
            None
        '''
        axe_max    = self.simplex.diagomax
        attrs      = self.simplex.attrs

        # We want a collection
        if not isinstance(observed_point, Collection):
            observed_point = Collection([observed_point])
         
        # Projection in barycentric coord.
        p3d = self.barycentric_to_3d(observed_point)
        # Supression of possible (None,None,None) 
        p3d = np.array( [ p for p in p3d if not None in p  ] )
        # Mean of cloud points
        (xp,yp,zp) = p3d.mean(axis=0)
        # Update camera_center
        attrs.camera_center = dict( x=(xp)/(2*axe_max), y=(yp)/(2*axe_max), z=(zp)/(2*axe_max) )


    def _get_layout(self):
        attrs     = self.simplex.attrs
        axe_max   = self.simplex.diagomax
        layout = dict( 
            width       = attrs.width, 
            height      = attrs.height,
            template    = attrs.plotly_template,
            autosize    = False,
            showlegend  = False,
            margin      = dict(l=0, r=0, t=0, b=0, pad=0),
            scene_xaxis = dict(nticks=10, range=[-axe_max, axe_max], visible=False),
            scene_yaxis = dict(nticks=10, range=[-axe_max, axe_max], visible=False),
            scene_zaxis = dict(nticks=10, range=[-axe_max, axe_max], visible=False),
            # scene_aspectratio   = dict( x=1, y=1, z=1 ),
            scene_aspectmode    = 'cube',
            scene_camera_eye    = attrs.camera_eye,
            scene_camera_center = attrs.camera_center
        )
        return layout

    def _get_config(self):
        attrs   = self.simplex.attrs
        config = {
            'toImageButtonOptions': {
                'format': 'png', 
                'filename': self.simplex.name,
                'width':  attrs.width,
                'height': attrs.height,
                'scale':  attrs.save_scale
            }
        }
        return config



    def _get_collection_labels(self,collection):
        lc = collection.labels
        ls = self.simplex.labels
        # ---- Labels still defined
        if lc is not None : return lc
        # ---- No labels
        labels=[]
        for p in collection.points:
            label=''
            for i,v in enumerate(p):
                label += f'{ls[i]} : {v}<br>'
            labels.append(label)
        return labels


    def _trace_collection(self, collection):
        attrs  = collection.attrs
        colors = collection.colors
        labels = self._get_collection_labels(collection)
        p      = self.barycentric_to_3d(collection)

        # ---- Markers
        #
        trace1 = go.Scatter3d(x=p[:, 0], y=p[:, 1], z=p[:, 2],
                              mode      = 'markers',
                              visible   = attrs.markers_visible,
                              hovertext = labels, 
                              hoverinfo = "text",
                              marker    = dict(color      = colors,
                                               opacity    = attrs.markers_opacity,
                                               colorscale = attrs.markers_colormap['colorscale'], 
                                               cmin       = attrs.markers_colormap['cmin'], 
                                               cmax       = attrs.markers_colormap['cmax'],
                                               size       = attrs.markers_size,
                                               line       = dict(width=attrs.markers_border_width, color = attrs.markers_border_color)))
        # ---- Lines
        #
        trace2 = go.Scatter3d(x=p[:, 0], y=p[:, 1], z=p[:, 2],
                              mode      = 'lines',
                              visible   = attrs.lines_visible,
                              opacity   = attrs.lines_opacity,
                              hovertext = [],
                              hoverinfo = 'skip',
                              line      = dict(color      = colors,
                                               colorscale = attrs.lines_colormap['colorscale'], 
                                               cmin       = attrs.lines_colormap['cmin'], 
                                               cmax       = attrs.lines_colormap['cmax'],
                                               width      = attrs.lines_width))
        # ---- Text
        #
        trace3 = go.Scatter3d(x=p[:, 0], y=p[:, 1], z=p[:, 2],
                              mode      = 'text',
                              visible   = attrs.text_visible,
                              opacity   = attrs.text_opacity,
                              hovertext = [],
                              hoverinfo = 'skip',
                              text      = labels,
                              textfont  = dict(size       = attrs.text_size))

        return [trace1, trace2, trace3]



    def plot(self, *stuffs, save_as=None, observed_point=None):

        attrs     = self.simplex.attrs

        # ---- Simplex Skeleton
        renderer = attrs.renderer
        skeleton = self.simplex.get_skeleton()
        traces   = self._trace_collection(skeleton)

        # ---- Add stuffs : collections as traces
        if stuffs is not None:
            for c in stuffs:
                traces.extend( self._trace_collection(c) )
                
        fig = go.Figure(data=[ *traces ])

        # ---- Observed point 
        if observed_point is not None:
            self.update_center(observed_point)

        # ---- Set layout
        fig.update_layout( self._get_layout() )

        # ---- Show it
        fig.show( config=self._get_config(), renderer=renderer )
        self.fig = fig

        # ---- Save option
        self.plot_save(save_as)


    def plot_save(self, save_as):
        attrs=self.simplex.attrs

        # ---- No save
        #
        if save_as is None : return

        # ---- Convert to a tuple if needed
        #
        if isinstance(save_as, str) : 
            save_as = (save_as,) 

        if not isinstance(save_as, Iterable) : 
            raise AttributeError('"save_as" must be str or list')

        # ---- Save it
        #
        for filename in save_as:
            if filename.endswith('.html') : 
                # ---- Save as html
                #
                fig=go.Figure(self.fig)
                fig.update_layout( dict(width=None, height=None, autosize=True) )
                pio.write_html(fig, filename)
            else:
                # ---- Save as an image
                #
                scale = attrs.save_scale
                size  = min(attrs.width, attrs.height)
                pio.write_image(self.fig, filename, width=size, height=size, scale=scale)
                
            print('Saved as : ', filename)







# ---------------------------------------------------------
#    ____          __      ___     
#   |  _ \         \ \    / (_)    
#   | |_) | __ _ _ _\ \  / / _ ____
#   |  _ < / _` | '__\ \/ / | |_  /
#   | |_) | (_| | |   \  /  | |/ / 
#   |____/ \__,_|_|    \/   |_/___|  Simplex class
# ---------------------------------------------------------                            
# SIMaP / Jean-Luc Parouty 2022    

import numpy as np
from math import sqrt, cos, sin, acos

from barviz.SimplexBuilder import SimplexBuilder
from barviz.Scrawler       import Scrawler
from barviz.Attributes     import Attributes
from barviz.Collection     import Collection
from barviz.version        import __version__

class Simplex(Collection):

    version = __version__
    type    = 'Simplex'

    _attributes_default= { 
        'markers_colormap'     : dict( colorscale='Turbo', cmin=0, cmax=None),
        'markers_opacity'      : 1,
        'markers_size'         : 8,
        'markers_visible'      : True,
        'markers_border_color' : 'darkgray',
        'markers_border_width' : 1,
        'lines_colormap'       : dict( colorscale='Turbo', cmin=0, cmax=None),
        'lines_opacity'        : 1,
        'lines_width'          : 2,
        'lines_visible'        : True,
        'text_opacity'         : 1,
        'text_size'            : 10,
        'text_visible'         : True,
        'width'                : 500,
        'height'               : 500,
        'camera_eye'           : dict(x=.8, y=.8, z=.8),
        'camera_center'        : dict(x=0, y=0, z=0),
        'save_scale'           : 6,
        'plotly_template'      : 'plotly_white',
        'renderer'             : None
    }

    def __init__( self,  points=[], name='unknown', colors=None, labels=None, attrs={} ):
        # ---- Super init
        super(Simplex, self).__init__(points,name,colors,labels,attrs)
        # ---- Simplex version and attributes
        self.version  = Simplex.version
        self._attrs   = Attributes(attrs, Simplex._attributes_default)
        # ---- Get a Scrawler for plotting
        self.scrawler = Scrawler(self)
        # ---- Add labels/colors if undefined
        if labels is None : self.labels=[ f'P{i}' for i in range(self.nbp) ]
        if colors is None : self.colors=[ i       for i in range(self.nbp) ]
        # ---- Add colormap cmax if undefined
        if self.attrs.markers_colormap['cmax'] is None : self.attrs.markers_colormap['cmax'] = self.nbp-1
        if self.attrs.lines_colormap['cmax']   is None : self.attrs.lines_colormap['cmax']   = self.nbp-1

    @classmethod
    def build(cls, nbp=5):
        spb = SimplexBuilder()
        spb.add_points(nbp)
        spb.iterate()
        return spb.get_simplex()


    def get_skeleton(self):
        n=self.nbp
        # ---- Vertices in barycentric world
        points=np.zeros( (n,n) )
        for i in range(n):
            points[i,i]=1

        # ---- Edges
        edges,colors,labels=[],[],[]
        for i in range(0,n):
            pi = points[i]
            for j in range(i+1,n):
                pj = points[j]
                edges.extend( (pi,pj,[None]*n) )
                colors.extend( (self.colors[i], self.colors[j], 0 ) )
                labels.extend( (self.labels[i], self.labels[j], '') )

        return Collection(points=edges,name='Skeleton',colors=colors,labels=labels,attrs=self.attrs.to_dict())

    def plot(self, *args, **kwargs):
        self.scrawler.plot(*args, **kwargs)

    def plot_save(self, *save_as):
        self.scrawler.plot_save(save_as)

    def resize(self, size=2):
        self.points = self.points * size / self.diagomax




# ---------------------------------------------------------
#    ____          __      ___     
#   |  _ \         \ \    / (_)    
#   | |_) | __ _ _ _\ \  / / _ ____
#   |  _ < / _` | '__\ \/ / | |_  /
#   | |_) | (_| | |   \  /  | |/ / 
#   |____/ \__,_|_|    \/   |_/___|  Attributes class
# ---------------------------------------------------------                            
# SIMaP / Jean-Luc Parouty 2022    

import json

class Attributes(object):

    def __init__(self, attributes, defaults={}):
        attrs=self.__dict__
        if isinstance(attributes, Attributes) :
            attributes=attributes.to_dict()
        for k,v in defaults.items():
            if isinstance(v,dict): v=dict(v)   # Need a copy !
            attrs[k]=attributes.get(k,v)

    def __getattr__(self, name):
        '''Get an attribute, if exist'''
        attrs=self.__dict__
        if name in attrs:
            return attrs[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        '''Set an attribute value, if it already exist'''
        attrs=self.__dict__
        if name in attrs:
            attrs[name]=value
        else:
            raise AttributeError("No such attribute: " + name)

    def to_dict(self):
        return self.__dict__

    def to_json(self):
        return json.dumps(self.__dict__, ensure_ascii=False, indent=4)

    def __repr__(self):
        attrs=self.__dict__
        out = 'Available attributes are :\n'
        for k,v in attrs.items():
            if not k.startswith('_') : out += f'    {k:18s} = {v}\n'
        return out



# ---------------------------------------------------------
#    ____          __      ___     
#   |  _ \         \ \    / (_)    
#   | |_) | __ _ _ _\ \  / / _ ____
#   |  _ < / _` | '__\ \/ / | |_  /
#   | |_) | (_| | |   \  /  | |/ / 
#   |____/ \__,_|_|    \/   |_/___|  Collection class
# ---------------------------------------------------------                            
# SIMaP / Jean-Luc Parouty 2022   

import numpy as np
import json
import pathlib
from barviz.Attributes     import Attributes
from barviz.version        import __version__


class Collection:

    version = __version__
    type    = 'Collection'

    _attributes_default= { 
        'markers_colormap'     : dict( colorscale=['gray','gray'], cmin=0, cmax=None),
        'markers_opacity'      : 0.6,
        'markers_size'         : 6,
        'markers_visible'      : True,
        'markers_border_color' : None,
        'markers_border_width' : 1,
        'lines_colormap'       : dict( colorscale=['gray','gray'], cmin=0, cmax=None),
        'lines_opacity'        : 1,
        'lines_width'          : 2,
        'lines_visible'        : False,
        'text_opacity'         : 1,
        'text_size'            : 10,
        'text_visible'         : False,
    }


    def __init__(self,  points=[], name='unknown', colors=None, labels=None, attrs={} ):
        self.version  = Collection.version
        self._attrs   = Attributes(attrs, Collection._attributes_default)
        self.name     = name
        self.points   = np.array(points)
        self.colors   = colors
        self.labels   = labels 
        # ---- Add labels/colors if undefined
        if labels is None : self.labels=[ f'C{i}' for i in range(self.nbp) ]
        if colors is None : self.colors=[ i       for i in range(self.nbp) ]
        # ---- Add colormap cmax if undefined
        if self.attrs.markers_colormap['cmax'] is None : self.attrs.markers_colormap['cmax'] = self.nbp-1
        if self.attrs.lines_colormap['cmax']   is None : self.attrs.lines_colormap['cmax']   = self.nbp-1


    def save_as(self, filename):
        # ---- Prepare data
        data = dict( 
            version   = self.version, 
            attrs     = self.attrs.to_dict(),
            name      = self.name,
            points    = self.points.tolist(),
            colors    = self.colors,
            labels    = self.labels
        )
        # ---- Create parents directories if needed
        pathlib.Path(filename).parent.mkdir(parents=True, exist_ok=True) 
        # ---- Save data as json
        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=4)
            print(f'Saved ({filename}).')


    @classmethod
    def load(cls, filename):
        # ---- Load json
        with open(filename) as jf:
            data = json.load(jf)
        # ---- Get a simplex
        sp = cls(points=data['points'], name=data['name'], colors=data['colors'], labels=data['labels'], attrs=data['attrs'])
        print(f'Loaded ({filename}).')
        return sp


    @property
    def attrs(self):
        return self._attrs

    @attrs.setter
    def attrs(self,attributes):
        self._attrs = Attributes(attributes, self._attributes_default)


    @property
    def nbp(self):
        return len(self.points)

    @property
    def diagomax(self):
        dm = -1
        for pi in self.points:
            if None not in pi : dm = max( dm, np.sqrt(pi.dot(pi)) )
        return dm


    def __repr__(self):
        return f'<Collection "{self.name}" with {self.nbp} points>'


    def __str__(self):
        p,c = self.points, self.colors
        out  = f'==== {self.type} : {self.name} ====\n'
        out += f'Version  : {self.version}\n'
        out += f'diagomax : {self.diagomax:.2f}\n' 
        out += f'Points   : {self.nbp}\n'
        for i,pi in enumerate(p[:25]):
            out += f'    {pi}   c={c[i]}\n'
        if self.nbp>25: out+='    (...)\n'
        return out

